from typing import Type, get_type_hints
import typing_inspect

from graphql.type.definition import \
    GraphQLType, \
    GraphQLObjectType, \
    GraphQLField, \
    GraphQLInputField, \
    GraphQLNonNull

from graphql_api.utils import to_camel_case


def type_is_dataclass(_class: Type):
    try:
        from dataclasses import is_dataclass
    except ImportError:
        return False
    else:
        return is_dataclass(_class)


def type_from_dataclass(_class: Type, mapper) -> GraphQLType:
    dataclass_fields = dict(_class.__dataclass_fields__)
    dataclass_types = get_type_hints(_class)
    base_type: GraphQLObjectType = mapper.map(_class, use_graphql_type=False)

    # Remove any modifiers
    while hasattr(base_type, 'of_type'):
        base_type = base_type.of_type

    if mapper.as_input:
        return base_type

    exclude_fields = _class.graphql_exclude_fields() \
        if hasattr(_class, 'graphql_exclude_fields') else []

    properties = {
        name: (field, dataclass_types.get(name))
        for name, field in dataclass_fields.items()
        if not name.startswith("_") and name not in exclude_fields
    }

    def local_fields_callback():
        local_type = base_type
        local_properties = properties
        local_mapper = mapper

        # noinspection PyProtectedMember
        local_type_fields = local_type._fields

        def fields_callback():
            local_fields = {}

            for prop_name, (field, field_type) in local_properties.items():

                def local_resolver():
                    local_prop_name = prop_name

                    def resolver(
                        self,
                        info=None,
                        context=None,
                        *args,
                        **kwargs
                    ):
                        return getattr(self, local_prop_name)
                    return resolver

                type_: GraphQLType = local_mapper.map(type_=field_type)

                nullable = False

                if typing_inspect.is_union_type(field_type):
                    union_args = typing_inspect.get_args(
                        field_type,
                        evaluate=True
                    )
                    if type(None) in union_args:
                        nullable = True

                if not nullable:
                    type_: GraphQLType = GraphQLNonNull(type_)

                if local_mapper.as_input:
                    field = GraphQLInputField(type_=type_)
                else:
                    field = GraphQLField(type_=type_, resolve=local_resolver())

                local_fields[to_camel_case(prop_name)] = field

            if local_type_fields:
                try:
                    fields_ = local_type_fields()
                    for name, field in fields_.items():
                        if name not in local_fields:
                            local_fields[name] = field
                except AssertionError:
                    pass

            return local_fields
        return fields_callback

    base_type._fields = local_fields_callback()
    return base_type



import datetime
import json
import uuid

from typing import Dict

from graphql import GraphQLScalarType, StringValueNode, Undefined
from graphql.language import ast


def parse_uuid_literal(ast):
    if isinstance(ast, StringValueNode):
        try:
            return uuid.UUID(ast.value)
        except ValueError:
            return Undefined


GraphQLUUID = GraphQLScalarType(
    name='UUID',
    description='The `UUID` scalar type represents a unique identifer.',
    serialize=str,
    parse_value=str,
    parse_literal=parse_uuid_literal)


def serialize_datetime(dt):
    return dt.isoformat(sep=" ")


def parse_datetime_value(value):
    datetime_formats = [
        "%Y-%m-%d %H:%M:%S.%f",
        "%Y-%m-%dT%H:%M:%S.%f"
    ]

    for datetime_format in datetime_formats:
        try:
            return datetime.datetime.strptime(value, datetime_format)
        except ValueError:
            pass

    raise ValueError(f"Datetime {value} did not fit any "
                     f"of the formats {datetime_formats}.")


def parse_datetime_literal(node):
    if isinstance(node, StringValueNode):
        return parse_datetime_value(node.value)


GraphQLDateTime = GraphQLScalarType(
    name='DateTime',
    description='The `DateTime` scalar type represents a datetime, '
                'the datetime should be in the format `2018-01-22 17:46:32`',
    serialize=serialize_datetime,
    parse_value=parse_datetime_value,
    parse_literal=parse_datetime_literal)


def serialize_json(data: Dict) -> str:
    return json.dumps(data)


def parse_json_value(value: str) -> Dict:
    return json.loads(value)


def parse_json_literal(node) -> Dict:
    if isinstance(node, ast.StringValueNode):
        return parse_json_value(node.value)


GraphQLJSON = GraphQLScalarType(
    name='JSON',
    description='The `JSON` scalar type represents JSON values as specified by'
                ' [ECMA-404](http://www.ecma-international.org/'
                'publications/files/ECMA-ST/ECMA-404.pdf).',
    serialize=serialize_json,
    parse_value=parse_json_value,
    parse_literal=parse_json_literal)


def serialize_bytes(bytes: bytes) -> str:
    return bytes.decode("utf-8")


def parse_bytes_value(value: str) -> bytes:
    return value.encode("utf-8")


def parse_bytes_literal(node):
    if isinstance(node, ast.StringValueNode):
        return parse_bytes_value(node.value)


GraphQLBytes = GraphQLScalarType(
    name='Bytes',
    description='The `Bytes` scalar type represents a '
                'Byte array datatype in UTF-8 string format.',
    serialize=serialize_bytes,
    parse_value=parse_bytes_value,
    parse_literal=parse_bytes_literal
)



import enum

from graphql import GraphQLObjectType, GraphQLNonNull

from graphql_api.context import GraphQLContext
from graphql_api.utils import to_snake_case


def middleware_local_proxy(next):
    value = next()

    # Compatibility with LocalProxy from Werkzeug
    try:
        if hasattr(value, '_get_current_object'):
            value = value._get_current_object()

    except Exception:
        pass

    if isinstance(value, Exception):
        raise value

    return value


def middleware_adapt_enum(next):
    """
    GraphQL middleware, by default enums return the value
    """
    value = next()
    if isinstance(value, enum.Enum):
        value = value.value

    return value


def middleware_request_context(next, context: GraphQLContext):
    from graphql_api.api import GraphQLRequestContext

    info = context.resolve_args.get('info')
    args = context.resolve_args.get('args')

    if info.context.request:
        return next()

    args = {to_snake_case(key): arg for key, arg in args.items()}
    graphql_request = GraphQLRequestContext(args=args, info=info)

    info.context.request = graphql_request

    try:
        value = next()
    finally:
        info.context.request = None

    return value


def middleware_field_context(next, context: GraphQLContext):
    from graphql_api.api import GraphQLFieldContext

    info = context.resolve_args.get('info')

    field_meta = info.context.meta.get(
        (info.parent_type.name, to_snake_case(info.field_name)),
        {}
    )
    return_type = info.return_type

    if field_meta is None:
        field_meta = {}

    if return_type and isinstance(return_type, GraphQLNonNull):
        return_type = return_type.of_type

    kwargs = {}
    if return_type and isinstance(return_type, GraphQLObjectType):
        sub_loc = info.field_nodes[0].selection_set.loc
        kwargs['query'] = sub_loc.source.body[sub_loc.start:sub_loc.end]

    info.context.field = GraphQLFieldContext(meta=field_meta, **kwargs)

    try:
        value = next()
    finally:
        info.context.field = None

    return value



from graphql_api.api import decorator


def field(meta=None, mutable=False):
    _type = "query"
    if mutable:
        _type = "mutation"

    return decorator(None, meta, _type=_type)


def type(meta=None, abstract=False, interface=False):
    _type = "object"
    if interface:
        _type = "interface"
    elif abstract:
        _type = "abstract"

    return decorator(None, meta, _type=_type)



import asyncio
import enum
import re
import aiohttp

from json.decoder import JSONDecodeError

from aiohttp import ClientTimeout
from graphql import (
    GraphQLNonNull,
    GraphQLList,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLError,
    build_client_schema,
    get_introspection_query)

from graphql.type.definition import GraphQLType, GraphQLInterfaceType


# From this response in Stackoverflow
# http://stackoverflow.com/a/19053800/1072990
def to_camel_case(snake_str, title=False):
    if snake_str.startswith('_'):
        snake_str = snake_str[1:]

    components = snake_str.split('_')
    # We capitalize the first letter of each component except the first one
    # with the 'title' method and join them together.
    if not snake_str:
        return ""
    prefix = components[0].title() if title else components[0]
    value = prefix + "".join(x.title() if x else '_' for x in components[1:])
    return value


# From this response in Stackoverflow
# http://stackoverflow.com/a/1176023/1072990
def to_snake_case(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def to_input_value(value):
    from graphql_api.mapper import is_scalar

    if value is None:
        return None

    python_type = type(value)

    if is_scalar(python_type):
        if isinstance(value, str):
            return '"' + value + '"'
        if isinstance(value, bool):
            return 'true' if value else 'false'
        else:
            return str(value)

    if isinstance(value, enum.Enum):
        return str(value.value)

    raise ValueError(f'Cannot map {value} to GraphQLInput')


def has_mutable(type, checked_types=None, interfaces_default_mutable=True):
    from .mapper import GraphQLMutableField, GraphQLTypeMapError

    while isinstance(type, (GraphQLNonNull, GraphQLList)):
        type = type.of_type

    if isinstance(type, (GraphQLObjectType, GraphQLInterfaceType)):
        if interfaces_default_mutable and \
                isinstance(type, GraphQLInterfaceType):
            return True

        if not checked_types:
            checked_types = set()
        try:
            fields = type.fields
        except (AssertionError, GraphQLTypeMapError):
            return False

        for key, field in fields.items():
            if isinstance(field, GraphQLMutableField):
                return True
            if field.type not in checked_types:
                checked_types.add(field.type)
                if has_mutable(
                    field.type,
                    checked_types,
                    interfaces_default_mutable
                ):
                    return True

    return False


def iterate_fields(type: GraphQLType, done_fields=None):
    from .mapper import GraphQLTypeMapError

    while isinstance(type, (GraphQLNonNull, GraphQLList)):
        type = type.of_type

    if isinstance(type, GraphQLObjectType):
        if not done_fields:
            done_fields = set()

        try:
            type.fields
        except (AssertionError, GraphQLTypeMapError):
            pass
        else:
            for key, field in type.fields.items():
                field_id = type.name + "." + key
                if field_id not in done_fields:
                    done_fields.add(field_id)
                    yield type, key, field
                    yield from iterate_fields(field.type, done_fields)


def url_to_ast(
    url,
    http_method="GET",
    http_headers=None,
    verify=True
) -> GraphQLSchema:
    _introspect_query = get_introspection_query()

    response = asyncio.run(
        http_query(
            url=url,
            query=_introspect_query,
            http_method=http_method,
            http_headers=http_headers,
            verify=verify
        )
    )
    errors = response.get('errors')

    if errors:
        raise GraphQLError(f"RemoteSchema {url} Error: {str(errors)}")

    introspect_schema = response.get('data')
    return build_client_schema(introspect_schema)


def executor_to_ast(executor) -> GraphQLSchema:
    _introspect_query = get_introspection_query()
    response = executor.execute(_introspect_query)
    introspect_schema = response.data

    return build_client_schema(introspect_schema)


async def http_query(
    url,
    query,
    variable_values=None,
    operation_name=None,
    http_method="GET",
    http_headers=None,
    http_timeout=10,
    verify=True
):
    params = {"query": query}

    if http_headers is None:
        http_headers = {}

    if variable_values:
        params["variables"] = variable_values

    if operation_name:
        params["operationName"] = operation_name

    async with aiohttp.ClientSession() as session:
        if http_method == "GET":
            r = await session.get(
                url,
                params=params,
                ssl=verify,
                headers={'Accept': 'application/json', **http_headers},
                timeout=ClientTimeout(total=http_timeout)
            )

        elif http_method == "POST":
            r = await session.post(
                url,
                json=params,
                ssl=verify,
                headers={'Accept': 'application/json', **http_headers},
                timeout=ClientTimeout(total=http_timeout)
            )

        else:
            raise AttributeError(f"Invalid HTTP method {http_method}")

        try:
            json = await r.json()
        except JSONDecodeError as e:
            raise ValueError(
                f"{e}, unable to decode JSON"
            )

    return json



from graphql import GraphQLError as GraphQLCoreError


class GraphQLError(GraphQLCoreError):

    def __init__(
        self,
        message,
        nodes=None,
        source=None,
        positions=None,
        path=None,
        original_error=None,
        extensions=None
    ):
        super(GraphQLError, self).__init__(
            message=message,
            nodes=nodes,
            source=source,
            positions=positions,
            path=path,
            original_error=original_error,
            extensions=extensions
        )
        self.extensions = extensions



from typing import List

from graphql import GraphQLNonNull, GraphQLList, GraphQLObjectType
from graphql.type.definition import GraphQLInterfaceType

from graphql_api.mapper import \
    GraphQLMutableField, \
    GraphQLTypeMapError, \
    GraphQLMetaKey
from graphql_api.utils import has_mutable, iterate_fields, to_snake_case


class GraphQLFilter:

    def filter_field(self, name, meta: dict) -> bool:
        """
        Return True to filter (remove) a field from the schema
        """
        raise NotImplementedError()


class TagFilter(GraphQLFilter):

    def __init__(self, tags: List[str] = None):
        """
        Remove any fields that are tagged with a tag in tags
        """
        self.tags = tags

    def filter_field(self, name: str, meta: dict) -> bool:
        tags = meta.get("tags", [])

        for tag in tags:
            if tag in self.tags:
                return True

        return False


class GraphQLSchemaReducer:

    @staticmethod
    def reduce_query(mapper, root, filters=None):
        query: GraphQLObjectType = mapper.map(root)

        # Remove any types that have no fields
        # (and remove any fields that returned that type)
        invalid_types, invalid_fields = GraphQLSchemaReducer.invalid(
            root_type=query,
            filters=filters,
            meta=mapper.meta
        )

        for type_, key in invalid_fields:
            del type_.fields[key]

        for key, value in dict(mapper.registry).items():
            if value in invalid_types:
                del mapper.registry[key]

        return query

    @staticmethod
    def reduce_mutation(mapper, root):
        mutation: GraphQLObjectType = mapper.map(root)

        # Trigger dynamic fields to be called
        for _ in iterate_fields(mutation):
            pass

        # Find all mutable Registry types
        filtered_mutation_types = {root}
        for type_ in mapper.types():
            if has_mutable(type_, interfaces_default_mutable=False):
                filtered_mutation_types.add(type_)

        # Replace fields that have no mutable
        # subtypes with their non-mutable equivalents

        for type_, key, field in iterate_fields(mutation):
            field_type = field.type
            meta = mapper.meta.get((type_.name, to_snake_case(key)), {})
            field_definition_type = meta.get('type', 'query')

            wraps = []
            while isinstance(field_type, (GraphQLNonNull, GraphQLList)):
                wraps.append(field_type.__class__)
                field_type = field_type.of_type

            if meta.get(GraphQLMetaKey.resolve_to_mutable):
                # Flagged as mutable
                continue

            if field_definition_type == "query":
                if mapper.suffix in str(field_type) or \
                        field_type in filtered_mutation_types:
                    # Calculated as it as mutable
                    continue

            # convert it to immutable
            query_type_name = str(field_type).replace(mapper.suffix, "", 1)
            query_type = mapper.registry.get(query_type_name)

            if query_type:
                for wrap in wraps:
                    query_type = wrap(query_type)
                field.type = query_type

        # Remove any query fields from mutable types
        fields_to_remove = set()
        for type_ in filtered_mutation_types:
            while isinstance(type_, (GraphQLNonNull, GraphQLList)):
                type_ = type_.of_type
            if isinstance(type_, GraphQLObjectType):
                interface_fields = []
                for interface in type_.interfaces:
                    interface_fields += [
                        key
                        for key, field in interface.fields.items()
                    ]
                for key, field in type_.fields.items():
                    if key not in interface_fields and \
                            not isinstance(field, GraphQLMutableField) and \
                            not has_mutable(field.type):
                        fields_to_remove.add((type_, key))

        for type_, key in fields_to_remove:
            del type_.fields[key]

        return mutation

    @staticmethod
    def invalid(
        root_type,
        filters=None,
        meta=None,
        checked_types=None,
        invalid_types=None,
        invalid_fields=None
    ):
        if not checked_types:
            checked_types = set()

        if not invalid_types:
            invalid_types = set()

        if not invalid_fields:
            invalid_fields = set()

        if root_type in checked_types:
            return invalid_types, invalid_fields

        checked_types.add(root_type)

        try:
            fields = root_type.fields
        except (AssertionError, GraphQLTypeMapError):
            invalid_types.add(root_type)
            return invalid_types, invalid_fields

        interfaces = []

        if hasattr(root_type, 'interfaces'):
            interfaces = root_type.interfaces

        interface_fields = []
        for interface in interfaces:
            try:
                interface_fields += [
                    key
                    for key, field in interface.fields.items()
                ]
            except (AssertionError, GraphQLTypeMapError):
                invalid_types.add(interface)

        for key, field in fields.items():
            if key not in interface_fields:
                type_ = field.type

                while isinstance(type_, (GraphQLNonNull, GraphQLList)):
                    type_ = type_.of_type

                field_name = to_snake_case(key)

                field_meta = meta.get((root_type.name, field_name), {})

                if filters:
                    for field_filter in filters:
                        if field_filter.filter_field(field_name, field_meta):
                            invalid_fields.add((root_type, key))

                if isinstance(
                    type_,
                    (GraphQLInterfaceType, GraphQLObjectType)
                ):
                    try:
                        assert type_.fields
                        sub_invalid = GraphQLSchemaReducer.invalid(
                            root_type=type_,
                            filters=filters,
                            meta=meta,
                            checked_types=checked_types,
                            invalid_types=invalid_types,
                            invalid_fields=invalid_fields
                        )

                        invalid_types.update(sub_invalid[0])
                        invalid_fields.update(sub_invalid[1])

                    except (AssertionError, GraphQLTypeMapError):
                        invalid_types.add(type_)
                        invalid_fields.add((root_type, key))

        return invalid_types, invalid_fields



from typing import List

from uuid import UUID, uuid4

from graphql_api.api import GraphQLAPI


@GraphQLAPI.type(interface=True)
class Node:
    """
    The `Node` Interface type represents a Relay Node.
    `https://facebook.github.io/relay/graphql/objectidentification.htm`
    """

    def __init__(self, id: UUID = None, *args, **kwargs):
        if id is None:
            id = uuid4()

        self.id = id
        super().__init__(*args, **kwargs)

    @property
    @GraphQLAPI.field
    def _id(self) -> UUID:
        return self.id


@GraphQLAPI.type
class PageInfo:
    """
    The `PageInfo` Object type represents a Relay PageInfo.
    `https://facebook.github.io/relay/graphql/connections.htm#sec-undefined.PageInfo`
    """

    def __init__(
        self,
        has_previous_page: bool,
        has_next_page: bool,
        start_cursor: str,
        end_cursor: str,
        *args,
        **kwargs
    ):

        super().__init__(*args, **kwargs)

        self._has_previous_page = has_previous_page
        self._has_next_page = has_next_page
        self._start_cursor = start_cursor
        self._end_cursor = end_cursor

    @property
    @GraphQLAPI.field
    def has_previous_page(self) -> bool:
        return self._has_previous_page

    @property
    @GraphQLAPI.field
    def has_next_page(self) -> bool:
        return self._has_next_page

    @property
    @GraphQLAPI.field
    def start_cursor(self) -> str:
        return self._start_cursor

    @property
    @GraphQLAPI.field
    def end_cursor(self) -> str:
        return self._end_cursor

    @property
    @GraphQLAPI.field
    def count(self) -> int:
        return self._count


@GraphQLAPI.type
class Edge:
    """
    The `Edge` Object type represents a Relay Edge.
    `https://facebook.github.io/relay/graphql/connections.htm#sec-Edge-Types`
    """

    def __init__(self, node: Node, cursor: str, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._node = node
        self._cursor = cursor

    @property
    @GraphQLAPI.field
    def node(self) -> Node:
        return self._node

    @property
    @GraphQLAPI.field
    def cursor(self) -> str:
        return self._cursor


@GraphQLAPI.type
class Connection:
    """
    The `Connection` Object type represents a Relay Connection.
    `https://facebook.github.io/relay/graphql/connections.htm#sec-Connection-Types`
    """

    def __init__(
        self,
        before: str = None,
        after: str = None,
        first: int = None,
        last: int = None, *args, **kwargs
    ):
        self._before = before
        self._after = after
        self._first = first
        self._last = last
        super().__init__(*args, **kwargs)

    @GraphQLAPI.field
    def edges(self) -> List[Edge]:
        raise NotImplementedError(
            f"{self.__class__.__name__} has not "
            f"implemented 'Connection.edges'"
        )

    @GraphQLAPI.field
    def page_info(self) -> PageInfo:
        raise NotImplementedError(
            f"{self.__class__.__name__} has not "
            f"implemented 'Connection.page_info'"
        )



import inspect

from typing import Any, List, Dict, Callable

from graphql import graphql, graphql_sync

from graphql.execution import ExecutionResult
from graphql.type.schema import GraphQLSchema

from graphql_api.context import GraphQLContext
from graphql_api.middleware import \
    middleware_field_context, \
    middleware_request_context, \
    middleware_local_proxy, \
    middleware_adapt_enum


class GraphQLBaseExecutor:

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.validate()

    def validate(self):
        pass

    def execute(
        self,
        query,
        variables=None,
        operation_name=None
    ) -> ExecutionResult:
        pass

    async def execute_async(
        self,
        query,
        variables=None,
        operation_name=None
    ) -> ExecutionResult:
        pass


class GraphQLExecutor(GraphQLBaseExecutor):

    def __init__(
        self,
        schema: GraphQLSchema,
        meta: Dict = None,
        root_value: Any = None,
        middleware: List[Callable[[Callable, GraphQLContext], Any]] = None,
        middleware_on_introspection: bool = False
    ):
        super().__init__()

        if meta is None:
            meta = {}

        if middleware is None:
            middleware = []

        middleware.insert(0, middleware_field_context)
        middleware.insert(0, middleware_request_context)
        middleware.insert(0, middleware_local_proxy)
        middleware.insert(0, middleware_adapt_enum)

        self.meta = meta
        self.schema = schema
        self.middleware = middleware
        self.root_value = root_value
        self.middleware_on_introspection = middleware_on_introspection

    def execute(
        self,
        query,
        variables=None,
        operation_name=None,
        root_value=None,
        context=None
    ) -> ExecutionResult:

        context = GraphQLContext(
            schema=self.schema,
            meta=self.meta,
            executor=self
        )

        if root_value is None:
            root_value = self.root_value

        value = graphql_sync(
            self.schema,
            query,
            context_value=context,
            variable_values=variables,
            operation_name=operation_name,
            middleware=self.adapt_middleware(self.middleware),
            root_value=root_value
        )
        return value

    async def execute_async(
        self,
        query,
        variables=None,
        operation_name=None,
        root_value=None,
        context=None
    ) -> ExecutionResult:

        context = GraphQLContext(
            schema=self.schema,
            meta=self.meta,
            executor=self
        )

        if root_value is None:
            root_value = self.root_value

        value = await graphql(
            self.schema,
            query,
            context_value=context,
            variable_values=variables,
            operation_name=operation_name,
            middleware=self.adapt_middleware(self.middleware),
            root_value=root_value
        )
        return value

    @staticmethod
    def adapt_middleware(
        middleware,
        middleware_on_introspection: bool = False
    ):

        def simplify(_middleware: Callable[[Callable, GraphQLContext], Any]):
            def graphql_middleware(next, root, info, **args):
                kwargs = {}
                if "context" in inspect.signature(_middleware).parameters:
                    context: GraphQLContext = info.context
                    kwargs["context"] = context
                    context.resolve_args['root'] = root
                    context.resolve_args['info'] = info
                    context.resolve_args['args'] = args

                return _middleware(lambda: next(root, info, **args), **kwargs)

            return graphql_middleware

        def skip_if_introspection(_middleware):
            def middleware_with_skip(next, root, info, **args):
                skip = info.operation.name and \
                       info.operation.name.value == 'IntrospectionQuery'
                if skip:
                    return next(root, info, **args)
                return _middleware(next, root, info, **args)

            return middleware_with_skip

        adapters = [simplify]

        if middleware_on_introspection is False:
            adapters.append(skip_if_introspection)

        adapted_middleware = []

        for middleware in reversed(middleware):
            for adapter in adapters:
                middleware = adapter(middleware)
            adapted_middleware.append(middleware)

        return adapted_middleware



import enum
import inspect
import collections.abc

import typing
import types
import typing_inspect

from uuid import UUID

from typing import List, Union, Type, Callable, Tuple, Any, Dict, Set

from typing_inspect import get_origin
from datetime import datetime

from graphql import (
    GraphQLObjectType,
    GraphQLField,
    GraphQLString,
    GraphQLList,
    GraphQLBoolean,
    GraphQLInt,
    GraphQLFloat
)

from graphql.type.definition import (
    GraphQLType,
    GraphQLUnionType,
    GraphQLInterfaceType,
    GraphQLArgument,
    GraphQLInputObjectType,
    is_input_type,
    GraphQLEnumType,
    GraphQLScalarType,
    GraphQLNonNull,
    GraphQLInputField
)

from graphql.pyutils import Undefined, UndefinedType

from graphql_api.context import GraphQLContext
from graphql_api.types import (
    GraphQLBytes,
    GraphQLUUID,
    GraphQLDateTime,
    GraphQLJSON
)

from graphql_api.utils import to_camel_case, to_snake_case, to_input_value
from graphql_api.exception import GraphQLBaseException
from graphql_api.dataclass_mapping import \
    type_is_dataclass, \
    type_from_dataclass

"""
class AnyObject:


    @classmethod
    def graphql_from_input(cls, age: int):
        pass

    # @classmethod
    # def graphql_fields(cls):
    #     pass

"""


class GraphQLTypeMapInvalid(GraphQLBaseException):
    pass


class GraphQLTypeMapError(GraphQLBaseException):
    pass


class GraphQLTypeWrapper:

    @classmethod
    def graphql_type(cls, mapper: "GraphQLTypeMapper") -> GraphQLType:
        pass


class GraphQLMetaKey(enum.Enum):
    resolve_to_mutable = "RESOLVE_TO_MUTABLE"
    resolve_to_self = "RESOLVE_TO_SELF"
    native_middleware = "NATIVE_MIDDLEWARE"


class GraphQLMutableField(GraphQLField):
    pass


class GraphQLTypeMapper:

    def __init__(
        self,
        as_mutable=False,
        as_input=False,
        registry=None,
        reverse_registry=None,
        suffix="",
        schema=None
    ):
        self.as_mutable = as_mutable
        self.as_input = as_input
        self.registry = registry or {}
        self.reverse_registry = reverse_registry or {}
        self.suffix = suffix
        self.meta = {}
        self.input_type_mapper = None
        self.schema = schema

    def types(self) -> Set[GraphQLType]:
        return set(self.registry.values())

    def map_to_field(
        self,
        function_type: Callable,
        name="",
        key=""
    ) -> GraphQLField:
        type_hints = typing.get_type_hints(function_type)
        description = inspect.getdoc(function_type)

        return_type = type_hints.pop('return', None)

        if not return_type:
            raise GraphQLTypeMapInvalid(
                f"Field '{name}.{key}' with function ({function_type}) did "
                f"not specify a valid return type."
            )

        return_graphql_type = self.map(return_type)

        nullable = False

        if typing_inspect.is_union_type(return_type):
            union_args = typing_inspect.get_args(return_type, evaluate=True)
            if type(None) in union_args:
                nullable = True

        if not self.validate(return_graphql_type):
            raise GraphQLTypeMapError(
                f"Field '{name}.{key}' with function '{function_type}' return "
                f"type '{return_type}' could not be mapped to a valid GraphQL "
                f"type, was mapped to invalid type {return_graphql_type}."
            )

        enum_return = None

        if isinstance(return_graphql_type, GraphQLEnumType):
            enum_return = return_type

        if not nullable:
            return_graphql_type: GraphQLType = GraphQLNonNull(
                return_graphql_type
            )

        signature = inspect.signature(function_type)

        default_args = {
            key: value.default
            for key, value in signature.parameters.items()
            if value.default is not inspect.Parameter.empty
        }

        input_type_mapper = GraphQLTypeMapper(
            as_mutable=self.as_mutable,
            as_input=True,
            registry=self.registry,
            reverse_registry=self.reverse_registry,
            suffix=self.suffix,
            schema=self.schema
        )
        self.input_type_mapper = input_type_mapper
        arguments = {}
        enum_arguments = {}

        include_context = False

        for key, hint in type_hints.items():
            if key == 'context' and \
                    inspect.isclass(hint) and \
                    issubclass(hint, GraphQLContext):
                include_context = True
                continue

            arg_type = input_type_mapper.map(hint)

            if isinstance(arg_type, GraphQLEnumType):
                enum_arguments[key] = hint

            nullable = key in default_args
            if not nullable:
                arg_type = GraphQLNonNull(arg_type)

            arguments[to_camel_case(key)] = GraphQLArgument(
                type_=arg_type,
                default_value=default_args.get(key, Undefined)
            )

        def resolve(self, info=None, context=None, *args, **kwargs):
            _args = {to_snake_case(key): arg for key, arg in kwargs.items()}

            if enum_arguments:
                enum_keys = list(enum_arguments.keys())
                _args = {
                    key: enum_arguments[key](arg)
                    if key in enum_keys else arg
                    for key, arg in _args.items()
                }

            if include_context:
                _args['context'] = info.context

            function_name = function_type.__name__
            parent_type = self.__class__
            class_attribute = getattr(parent_type, function_name, None)
            is_property = isinstance(class_attribute, property)
            response = None

            if is_property:
                if _args:
                    if len(_args) > 1:
                        raise KeyError(
                            f"{function_name} on type {parent_type} is a"
                            f" property, and cannot have multiple arguments."
                        )
                    else:
                        response = function_type(self, **_args)
                else:
                    response = getattr(self, function_name, None)
            else:
                function_type_override = getattr(self, function_name, None)

                if function_type_override is not None:
                    response = function_type_override(**_args)
                else:
                    response = function_type(self, **_args)

            if enum_return:
                if isinstance(response, enum.Enum):
                    response = response.value

            return response

        field_class = GraphQLField
        func_type = get_value(function_type, self.schema, 'type')
        if func_type == "mutation":
            field_class = GraphQLMutableField

        return field_class(
            return_graphql_type,
            arguments,
            resolve,
            description=description
        )

    def map_to_union(self, union_type: Union) -> GraphQLType:
        union_args = typing_inspect.get_args(union_type, evaluate=True)
        none_type = type(None)
        union_map: Dict[type, GraphQLType] = {
            arg: self.map(arg)
            for arg in union_args if arg and arg != none_type
        }

        if len(union_map) == 1:
            _, mapped_type = union_map.popitem()
            return mapped_type

        def resolve_type(value, info, _type):
            from graphql_api.remote import GraphQLRemoteObject

            value_type = type(value)

            if isinstance(value, GraphQLRemoteObject):
                value_type = value.python_type

            for arg, mapped_type in union_map.items():
                if issubclass(value_type, arg):
                    return mapped_type.name

        names = [arg.__name__ for arg in union_args]
        name = f"{''.join(names)}{self.suffix}Union"

        return GraphQLUnionType(
            name,
            types=[*union_map.values()],
            resolve_type=resolve_type
        )

    def map_to_list(self, type_: List) -> GraphQLList:
        list_subtype = typing_inspect.get_args(type_)[0]

        list_type = GraphQLList(type_=self.map(list_subtype))

        return list_type

    def map_to_enum(self, type_: Type[enum.Enum]) -> GraphQLEnumType:
        enum_type = type_
        name = f"{type_.__name__}Enum"
        # Enums dont include a suffix as they are immutable

        description = inspect.getdoc(enum_type)

        enum_type = GraphQLEnumType(
            name=name,
            values=enum_type,
            description=description
        )

        enum_type.enum_type = type_

        def serialize(self, value) -> Union[str, None, UndefinedType]:
            if value and isinstance(value, collections.abc.Hashable):
                if isinstance(value, enum.Enum):
                    value = value.value

                lookup_value = self._value_lookup.get(value)
                if lookup_value:
                    return lookup_value
                else:
                    return Undefined

            return None

        enum_type.serialize = types.MethodType(serialize, enum_type)

        return enum_type

    scalar_map = [
        ([UUID], GraphQLUUID),
        ([str], GraphQLString),
        ([bytes], GraphQLBytes),
        ([bool], GraphQLBoolean),
        ([int], GraphQLInt),
        ([dict, list, set], GraphQLJSON),
        ([float], GraphQLFloat),
        ([datetime], GraphQLDateTime),
        ([type(None)], None)
    ]

    def scalar_classes(self):
        classes = []
        for scalar_class_map in self.scalar_map:
            for scalar_class in scalar_class_map[0]:
                classes.append(scalar_class)
        return classes

    def map_to_scalar(self, class_type: Type) -> GraphQLScalarType:
        for test_types, graphql_type in self.scalar_map:
            for test_type in test_types:
                if issubclass(class_type, test_type):
                    return graphql_type

    def map_to_interface(self, class_type: Type, ) -> GraphQLType:
        subclasses = class_type.__subclasses__()
        name = class_type.__name__

        for subclass in subclasses:
            if not is_abstract(subclass, self.schema):
                self.map(subclass)

        class_funcs = get_class_funcs(class_type, self.schema, self.as_mutable)

        interface_name = f"{name}{self.suffix}Interface"
        description = inspect.getdoc(class_type)

        def local_resolve_type():
            local_self = self

            def resolve_type(value, info, _type):
                return local_self.map(type(value)).name
            return resolve_type

        def local_fields():
            local_self = self
            local_class_funcs = class_funcs
            local_class_type = class_type
            local_name = name

            def fields():
                fields_ = {}
                for key_, func_ in local_class_funcs:
                    local_class_name = local_class_type.__name__
                    func_.__globals__[local_class_name] = local_class_type
                    fields_[to_camel_case(key_)] = local_self.map_to_field(
                        func_,
                        local_name,
                        key_
                    )

                return fields_
            return fields

        return GraphQLInterfaceType(
            interface_name,
            fields=local_fields(),
            resolve_type=local_resolve_type(),
            description=description
        )

    def map_to_input(self, class_type: Type) -> GraphQLType:
        name = f"{class_type.__name__}{self.suffix}Input"

        if hasattr(class_type, 'graphql_from_input'):
            creator = class_type.graphql_from_input
            func = creator

        else:
            creator = class_type
            func = class_type.__init__

        description = inspect.getdoc(func) or inspect.getdoc(class_type)

        try:
            type_hints = typing.get_type_hints(func)
        except Exception as err:
            raise TypeError(
                f"Unable to build input type '{name}' for '{class_type}', "
                f"check the '{class_type}.__init__' method or the "
                f"'{class_type}.graphql_from_input' method, {err}."
            )
        type_hints.pop("return", None)

        signature = inspect.signature(func)

        default_args = {
            key: value.default
            for key, value in signature.parameters.items()
            if value.default is not inspect.Parameter.empty
        }

        def local_fields():

            local_name = name
            local_self = self
            local_type_hints = type_hints
            local_default_args = default_args

            def fields():
                arguments = {}

                for key, hint in local_type_hints.items():

                    input_arg_type = local_self.map(hint)

                    nullable = key in local_default_args
                    if not nullable:
                        input_arg_type = GraphQLNonNull(input_arg_type)

                    default_value = local_default_args.get(key, None)

                    if default_value is not None:
                        try:
                            default_value = to_input_value(default_value)
                        except ValueError as err:
                            raise ValueError(
                                f"Unable to map {local_name}.{key}, {err}."
                            )

                    arguments[to_camel_case(key)] = GraphQLInputField(
                        type_=input_arg_type,
                        default_value=default_value
                    )
                return arguments

            return fields

        def local_container_type():
            local_creator = creator

            def container_type(data):
                data = {
                    to_snake_case(key): value
                    for key, value in data.items()
                }
                return local_creator(**data)

            return container_type

        return GraphQLInputObjectType(
            name,
            fields=local_fields(),
            out_type=local_container_type(),
            description=description
        )

    def map_to_object(self, class_type: Type) -> GraphQLType:
        name = f"{class_type.__name__}{self.suffix}"
        description = inspect.getdoc(class_type)

        class_funcs = get_class_funcs(class_type, self.schema, self.as_mutable)

        for key, func in class_funcs:
            func_meta = get_value(func, self.schema, 'meta')
            func_meta['type'] = get_value(func, self.schema, 'type')

            self.meta[(name, to_snake_case(key))] = func_meta

        def local_interfaces():
            local_class_type = class_type
            local_self = self

            def interfaces():
                _interfaces = []
                superclasses = inspect.getmro(local_class_type)[1:]

                for superclass in superclasses:
                    if is_interface(superclass, local_self.schema):
                        _interfaces.append(local_self.map(superclass))

                return _interfaces

            return interfaces

        def local_fields():
            local_self = self
            local_class_funcs = class_funcs
            local_class_type = class_type
            local_name = name

            def fields():
                fields_ = {}

                for key_, func_ in local_class_funcs:
                    local_class_type_name = local_class_type.__name__
                    func_.__globals__[local_class_type_name] = local_class_type
                    fields_[to_camel_case(key_)] = local_self.map_to_field(
                        func_,
                        local_name,
                        key_
                    )

                return fields_

            return fields

        obj = GraphQLObjectType(
            name,
            local_fields(),
            local_interfaces(),
            description=description
        )

        return obj

    def rmap(self, graphql_type: GraphQLType) -> Type:
        while hasattr(graphql_type, 'of_type'):
            graphql_type = graphql_type.of_type

        return self.reverse_registry.get(graphql_type)

    def map(self, type_, use_graphql_type=True) -> GraphQLType:

        def _map(type__) -> GraphQLType:

            if use_graphql_type and inspect.isclass(type__):
                if issubclass(type__, GraphQLTypeWrapper):
                    return type__.graphql_type(mapper=self)

                if type_is_dataclass(type__):
                    return type_from_dataclass(type__, mapper=self)

            if typing_inspect.is_union_type(type__):
                return self.map_to_union(type__)

            origin_type = get_origin(type__)

            if inspect.isclass(origin_type) and \
                    issubclass(get_origin(type__), (List, Set)):
                return self.map_to_list(type__)

            if inspect.isclass(type__):
                if issubclass(type__, GraphQLType):
                    return type__()

                if issubclass(type__, tuple(self.scalar_classes())):
                    return self.map_to_scalar(type__)

                if issubclass(type__, enum.Enum):
                    return self.map_to_enum(type__)

                if is_interface(type__, self.schema):
                    return self.map_to_interface(type__)

                if self.as_input:
                    return self.map_to_input(type__)
                else:
                    return self.map_to_object(type__)

            if isinstance(type__, GraphQLType):
                return type__

        key_hash = abs(hash(str(type_))) % (10 ** 8)
        suffix = {'|' + self.suffix if self.suffix else ''}
        generic_key = f"Registry({key_hash})" \
                      f"{suffix}|{self.as_input}|{self.as_mutable}"

        generic_registry_value = self.registry.get(generic_key, None)

        if generic_registry_value:
            return generic_registry_value

        value: GraphQLType = _map(type_)
        key = str(value)

        registry_value = self.registry.get(key, None)

        if not registry_value:
            self.register(python_type=type_, key=key, value=value)
            self.register(python_type=type_, key=generic_key, value=value)
            return value

        return registry_value

    def register(self, python_type: Type, key: str, value: GraphQLType):
        if self.validate(value):
            self.registry[key] = value
            self.reverse_registry[value] = python_type

    def validate(self, type_: GraphQLType, evaluate=False) -> bool:
        if not type_:
            return False

        if not isinstance(type_, GraphQLType):
            return False

        if isinstance(type_, GraphQLNonNull):
            type_ = type_.of_type

        if self.as_input and not is_input_type(type_):
            return False

        if isinstance(type_, GraphQLObjectType):
            if evaluate:
                try:
                    if len(type_.fields) == 0:
                        return False
                except AssertionError:
                    return False

            elif not callable(type_._fields) and len(type_._fields) == 0:
                return False

        return True


def get_class_funcs(
    class_type,
    schema,
    mutable=False
) -> List[Tuple[Any, Any]]:
    members = [(key, member) for key, member in inspect.getmembers(class_type)]

    if hasattr(class_type, 'graphql_fields'):
        members += [
            (func.__name__, func)
            for func in class_type.graphql_fields()
        ]
    func_members = []

    for key, member in members:
        if isinstance(member, property):
            getter = member.fget
            if getter:
                func_members.append((key, getter))
            setter = member.fset

            if setter:
                func_members.append((key, setter))
        else:
            func_members.append((key, member))

    def matches_criterion(func):
        func_type = get_value(func, schema, 'type')
        return func_type == "query" or (mutable and func_type == "mutation")

    callable_funcs = []

    for key, member in func_members:
        if is_graphql(member, schema=schema) and matches_criterion(member):
            if not callable(member):
                type_hints = typing.get_type_hints(member)
                return_type = type_hints.pop('return', None)

                def local_func():
                    local_key = key
                    local_member = member

                    def func(self) -> return_type:
                        return getattr(self, local_key)

                    func.graphql = local_member.graphql
                    func.defined_on = local_member.defined_on
                    func.schemas = {
                        schema: {
                            "meta": local_member.meta,
                            "type": local_member.type,
                            "defined_on": local_member.defined_on,
                            "schema": schema
                        }
                    }

                    return func

                func = local_func()

            else:
                func = member

            callable_funcs.append((key, func))

    return callable_funcs


def get_value(type_, schema, key):
    if is_graphql(type_, schema):
        return type_.schemas.get(schema, type_.schemas.get(None)).get(key)


def is_graphql(type_, schema):
    graphql = getattr(type_, 'graphql', None)
    schemas = getattr(type_, 'schemas', {})

    valid_schema = schema in schemas.keys() or None in schemas.keys()

    return graphql and schemas and valid_schema


def is_interface(type_, schema):
    if is_graphql(type_, schema):
        type_type = get_value(type_, schema, 'type')
        type_defined_on = get_value(type_, schema, 'defined_on')

        return type_type == "interface" and type_defined_on == type_


def is_abstract(type_, schema):
    if is_graphql(type_, schema):
        type_type = get_value(type_, schema, 'type')
        type_defined_on = get_value(type_, schema, 'defined_on')

        return type_type == "abstract" and type_defined_on == type_


def is_scalar(type_):
    for test_types, graphql_type in GraphQLTypeMapper.scalar_map:
        for test_type in test_types:
            if issubclass(type_, test_type):
                return True
    return False



class GraphQLContext:

    def __init__(self, schema=None, meta=None, executor=None):

        if not meta:
            meta = {}

        # Server specific
        self.schema = schema
        self.meta = meta
        self.executor = executor

        # Request specific
        self.request = None

        # Request field specific
        self.field = None

        # resolve arg data
        self.resolve_args = {}

    def __copy__(self):
        context_copy = type(self)()
        context_copy.__dict__.update(self.__dict__)
        return context_copy



import io
import json
import urllib.request

from setuptools import setup, find_packages

with io.open('README.md', 'rt', encoding='utf8') as readme_file:
    readme = readme_file.read()


def current_pypi_version(package):
    r = urllib.request.urlopen(f'https://pypi.python.org/pypi/{package}/json')
    if r.code == 200:
        return json.loads(r.read()).get('info').get('version')


def increment_ver(version):
    version = version.split('.')
    version[2] = str(int(version[2]) + 1)
    return '.'.join(version)


version = increment_ver(current_pypi_version('graphql-api'))

setup(
    name='graphql-api',
    version=version,
    license='MIT',
    packages=find_packages(),
    include_package_data=True,
    author='Robert Parker',
    author_email='rob@parob.com',
    url='https://gitlab.com/parob/graphql-api',
    download_url=f'https://gitlab.com/parob/graphql/-/archive/v{version}/graphql-api-v{version}.tar.gz',
    keywords=['GraphQL', 'GraphQL-API', 'GraphQLAPI', 'Server'],
    description='A framework for building Python GraphQL APIs.',
    long_description=readme,
    long_description_content_type='text/markdown',
    install_requires=[
        'graphql-core>=3.1.2',
        'requests>=2.24.0',
        'typing-inspect>=0.6.0',
        'aiohttp>=3.8.0'
    ],
    extras_require={
        'dev': [
            'pytest>=5.4.3',
            'pytest-cov>=2.10.0',
            'coverage>=5.2',
            'faker>=4.1.1'
        ]
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import asyncio
import enum
import inspect
import json
import uuid

from typing import List, Tuple, Dict, Type

from graphql.language import ast
from requests.exceptions import RequestException

from graphql import (
    GraphQLInputObjectType,
    GraphQLObjectType,
    GraphQLEnumType,
    GraphQLInterfaceType,
    GraphQLUnionType,
    GraphQLID,
    GraphQLString,
    GraphQLFloat,
    GraphQLBoolean,
    GraphQLInt
)
from graphql.execution import ExecutionResult
from graphql.type.definition import (
    GraphQLField,
    GraphQLScalarType,
    GraphQLNonNull,
    GraphQLList,
    GraphQLType,
    is_enum_type
)

from graphql_api.error import GraphQLError
from graphql_api.executor import GraphQLBaseExecutor
from graphql_api.mapper import GraphQLTypeMapper, GraphQLMetaKey
from graphql_api.api import GraphQLAPI
from graphql_api.utils import \
    to_camel_case, \
    url_to_ast, \
    to_snake_case, \
    http_query


class GraphQLRemoteExecutor(GraphQLBaseExecutor, GraphQLObjectType):

    def __init__(
        self,
        url,
        name="Remote",
        description=None,
        http_method="GET",
        http_headers=None,
        http_timeout=None,
        verify=True,
        ignore_unsupported=True
    ):

        if not description:
            description = f'The `{name}` object type forwards all ' \
                          f'requests to the GraphQL executor at {url}'

        if http_headers is None:
            http_headers = {}

        self.url = url
        self.http_method = http_method
        self.http_headers = http_headers
        self.http_timeout = http_timeout
        self.verify = verify
        self.ignore_unsupported = ignore_unsupported

        super().__init__(name=name,
                         fields=self.build_fields,
                         description=description)

    def build_fields(self):
        ast_schema = url_to_ast(
            self.url,
            http_method=self.http_method,
            http_headers=self.http_headers
        )

        def resolver(info=None, context=None, *args, **kwargs):
            field_ = context.field_nodes[0]
            if field_.alias:
                key_ = field_.alias.value
            else:
                key_ = field_.name.value

            return info[key_]

        # noinspection PyProtectedMember
        for name, type in ast_schema.type_map.items():
            if (isinstance(type, GraphQLObjectType) or
                isinstance(type, GraphQLInputObjectType)) and \
                    not type.name.startswith("__"):
                for key, field in type.fields.items():
                    field.resolver = resolver
            elif isinstance(type, GraphQLEnumType):
                if not self.ignore_unsupported:
                    raise GraphQLError(
                        f"GraphQLScalarType '{type}' type is not supported "
                        f"in a remote executor '{self.url}'."
                    )
            elif isinstance(type, (GraphQLInterfaceType, GraphQLUnionType)):

                super_type = 'GraphQLInterface' \
                    if isinstance(type, GraphQLInterfaceType) \
                    else 'GraphQLUnionType'

                if not self.ignore_unsupported:
                    raise GraphQLError(
                        f"{super_type} '{type}' type is not supported"
                        f" from remote executor '{self.url}'."
                    )
            elif isinstance(type, GraphQLScalarType):
                if type not in [
                    GraphQLID,
                    GraphQLString,
                    GraphQLFloat,
                    GraphQLBoolean,
                    GraphQLInt
                ]:
                    if not self.ignore_unsupported:
                        raise GraphQLError(
                            f"GraphQLScalarType '{type}' type is not "
                            f"supported in a remote executor '{self.url}'."
                        )
            elif str(type).startswith('__'):
                continue
            else:
                raise GraphQLError(
                    f"Unknown GraphQLType '{type}' type is not supported in "
                    f"a remote executor '{self.url}'."
                )

        # noinspection PyProtectedMember
        return ast_schema.query_type.fields

    async def execute_async(
        self,
        query,
        variable_values=None,
        operation_name=None,
        http_headers=None
    ) -> ExecutionResult:

        if http_headers is None:
            http_headers = self.http_headers
        else:
            http_headers = {**self.http_headers, **http_headers}

        try:
            json_ = await http_query(
                url=self.url,
                query=query,
                variable_values=variable_values,
                operation_name=operation_name,
                http_method=self.http_method,
                http_headers=http_headers,
                http_timeout=self.http_timeout,
                verify=self.verify
            )
        except RequestException as e:
            import sys
            err_msg = f"{e}, remote service '{self.name}' is unavailable."
            raise type(e)(err_msg).with_traceback(sys.exc_info()[2])

        except ValueError as e:
            raise ValueError(
                f"{e}, from remote service '{self.name}'."
            )

        return ExecutionResult(
            data=json_.get('data'),
            errors=json_.get('errors')
        )

    def execute(
        self,
        query,
        variable_values=None,
        operation_name=None,
        http_headers=None
    ) -> ExecutionResult:

        if http_headers is None:
            http_headers = self.http_headers
        else:
            http_headers = {**self.http_headers, **http_headers}

        try:
            json_ = asyncio.run(http_query(
                url=self.url,
                query=query,
                variable_values=variable_values,
                operation_name=operation_name,
                http_method=self.http_method,
                http_headers=http_headers,
                http_timeout=self.http_timeout,
                verify=self.verify
            ))
        except RequestException as e:
            import sys
            err_msg = f"{e}, remote service '{self.name}' is unavailable."
            raise type(e)(err_msg).with_traceback(sys.exc_info()[2])

        except ValueError as e:
            raise ValueError(
                f"{e}, from remote service '{self.name}'."
            )

        return ExecutionResult(
            data=json_.get('data'),
            errors=json_.get('errors')
        )


class GraphQLMappers:

    def __init__(
        self,
        query_mapper: GraphQLTypeMapper,
        mutable_mapper: GraphQLTypeMapper
    ):
        self.query_mapper = query_mapper
        self.mutable_mapper = mutable_mapper

    def map(self, type, reverse=False):
        query_type = None
        try:
            if reverse:
                query_type = self.query_mapper.rmap(type)
            else:
                query_type = self.query_mapper.map(type)
        except Exception:
            pass

        mutable_type = None
        try:
            if reverse:
                mutable_type = self.mutable_mapper.rmap(type)
            else:
                mutable_type = self.mutable_mapper.map(type)
        except Exception:
            pass

        if reverse:
            return query_type or mutable_type

        return query_type, mutable_type


class NullResponse(BaseException):
    pass


class GraphQLRemoteError(GraphQLError):

    def __init__(self, query=None, result=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.query = query
        self.result = result


class GraphQLAsyncStub:

    async def call_async(self, name, *args, **kwargs):
        pass


class GraphQLRemoteObject:

    def get_labels(self) -> List[str]:
        return self.python_type.get_labels()

    @classmethod
    def from_url(
        cls,
        url: str,
        api: GraphQLAPI,
        http_method: str = "GET"
    ) -> 'GraphQLRemoteObject':
        executor = GraphQLRemoteExecutor(url=url, http_method=http_method)

        return GraphQLRemoteObject(executor=executor, api=api)

    # noinspection PyProtectedMember
    def __init__(
        self,
        executor: GraphQLBaseExecutor,
        api: GraphQLAPI = None,
        mappers: GraphQLMappers = None,
        python_type: Type = None,
        call_history: List[Tuple['GraphQLRemoteField', Dict]] = None,
        delay_mapping: bool = True
    ):
        if not call_history:
            call_history = []

        if not api and python_type:
            api = GraphQLAPI(root=python_type)

        elif not python_type:
            python_type = api.root_type

        self.executor = executor
        self.api = api
        self.mappers = mappers
        self.call_history = call_history
        self.values = {}
        self.python_type = python_type
        self.mapped_types = False
        self.graphql_query_type = None
        self.graphql_mutable_type = None

        if not delay_mapping:
            self._map()

    def clear_cache(self):
        self.values.clear()

    def _map(self, force=False):
        if self.mappers is None:
            api = self.api

            api.graphql_schema()

            self.mappers = GraphQLMappers(
                query_mapper=api.query_mapper,
                mutable_mapper=api.mutation_mapper
            )

        if not self.mapped_types:
            self.mapped_types = True
            graphql_types = self.mappers.map(self.python_type)
            self.graphql_query_type, self.graphql_mutable_type = graphql_types

    def fetch(self, fields: List[Tuple['GraphQLRemoteField', Dict]] = None):
        if fields is None:
            fields = self._fields()

        field_values = self._fetch(fields=fields)

        for field, args in fields:
            field_value = field_values.get(to_camel_case(field.name))

            arg_hash = self.hash(args)

            self.values[(field, arg_hash)] = field_value

    async def fetch_async(
        self,
        fields: List[Tuple['GraphQLRemoteField', Dict]] = None
    ):
        if fields is None:
            fields = self._fields()

        field_values = await self._fetch_async(fields=fields)

        for field, args in fields:
            field_value = field_values.get(to_camel_case(field.name))

            arg_hash = self.hash(args)

            self.values[(field, arg_hash)] = field_value

    def _fields(self):
        self._map()

        def is_valid_field(field):
            if not is_scalar(field.type):
                return False

            for arg in field.args.values():
                if isinstance(arg.type, GraphQLNonNull):
                    return False

            return True

        valid_fields = [
            name
            for name, field in self.graphql_query_type.fields.items()
            if is_valid_field(field)
        ]

        return [(self.get_field(name), {}) for name in valid_fields]

    def _fetch(self, fields: List[Tuple['GraphQLRemoteField', Dict]] = None):
        """
        Load all the scalar values for this object into the values dictionary
        :return:
        """
        if fields is None:
            fields = self._fields()

        query = self._fetch_build_query(fields=fields)

        result = self.executor.execute(query=query)

        return self._fetch_process(query, result, fields)

    async def _fetch_async(
        self,
        fields: List[Tuple['GraphQLRemoteField', Dict]] = None
    ):
        """
        Load all the scalar values for this object into the values dictionary
        :return:
        """
        if fields is None:
            fields = self._fields()

        query = self._fetch_build_query(fields=fields)

        result = await self.executor.execute_async(query=query)

        return self._fetch_process(query, result, fields)

    def _fetch_build_query(
        self,
        fields: List[Tuple['GraphQLRemoteField', Dict]]
    ):
        self._map()

        mutable = any([
            field.mutable
            for field, args in self.call_history + fields])

        query_builder = GraphQLRemoteQueryBuilder(
            call_stack=self.call_history,
            fields=fields,
            mappers=self.mappers,
            mutable=mutable
        )

        return query_builder.build()

    def _fetch_process(
        self,
        query,
        result: ExecutionResult,
        fields: List[Tuple['GraphQLRemoteField', Dict]]
    ):

        if result.errors:
            raise GraphQLRemoteError(
                query=query,
                result=result,
                message=result.errors[0]['message']
            )

        field_values = result.data
        for field, args in self.call_history:
            camel_name = to_camel_case(field.name)

            if isinstance(field_values, list):
                raise ValueError(
                    "GraphQLLists can only contain scalar values."
                )

            if field_values is None:
                raise NullResponse()

            field_values = field_values.get(camel_name)

        if field_values is None:
            raise NullResponse()

        def parse_field(key, value):

            field: GraphQLRemoteField = None

            for _field, _field_dict in fields:
                if _field.name == key:
                    field = _field
                    break

            if not field:
                raise KeyError(
                    f"Could not find field for key {key}"
                )

            field_type = field.graphql_type()

            if value is None:
                if not field.nullable:
                    raise TypeError(
                        f"Unable to parse None type for non nullable field, "
                        f"'{key}'. Expected type: {field_type}"
                    )
                return None

            if not is_scalar(field_type):
                raise TypeError(
                    f"Unable to parse non-scalar type {field_type}"
                )

            ast_value = to_ast_value(value, field_type)

            if hasattr(field_type, 'parse_literal'):
                value = field_type.parse_literal(ast_value)

                if is_enum_type(field_type) and \
                        hasattr(field_type, 'enum_type'):
                    enum_type = field_type.enum_type
                    value = enum_type(value)

                return value

            raise TypeError(
                f"Scalar type {field_type} missing 'parse_literal' attribute"
            )

        if isinstance(field_values, list):
            field_values = [
                {
                    key: parse_field(key, value)
                    for key, value in field_values_list_item.items()
                }
                for field_values_list_item in field_values
            ]
        else:
            field_values = {
                key: parse_field(key, value)
                for key, value in field_values.items()
            }

        return field_values

    def hash(self, args: Dict):
        hashable_args = {}

        for key, value in args.items():

            if isinstance(value, list):
                value = tuple(value)

            hashable_args[key] = value

        return hash(frozenset(hashable_args.items()))

    def _get_value_cached(self, field: 'GraphQLRemoteField', args: Dict):
        try:
            arg_hash = self.hash(args)
        except TypeError:
            arg_hash = hash(uuid.uuid4())

        if field.mutable:
            self.values.clear()

        for ((_field, _arg_hash), value) in self.values.items():
            if field.name == _field.name and arg_hash == _arg_hash:
                return value, True, arg_hash

        return None, False, arg_hash

    def _get_value_check_mutated(self, field):
        mutated = any([field.mutable for field, args in self.call_history])

        if mutated and (field.scalar or field.mutable or field.nullable):
            raise GraphQLError(
                f"Cannot fetch {field.name} from {self.python_type}, "
                f"mutated objects cannot be re-fetched."
            )

    async def get_value_async(self, field: 'GraphQLRemoteField', args: Dict):
        self._map()
        value, result, arg_hash = self._get_value_cached(field, args)

        if result:
            return value

        if (field, arg_hash) in self.values.keys():
            return self.values.get((field, arg_hash), None)

        self._get_value_check_mutated(field)

        if field.scalar:
            await self.fetch_async(fields=[(field, args)])
            return self.values.get((field, arg_hash), None)

        else:
            python_type = self.mappers.map(
                field.graphql_field.type,
                reverse=True
            )

            obj = GraphQLRemoteObject(
                executor=self.executor,
                api=self.api,
                python_type=python_type,
                mappers=self.mappers,
                call_history=[*self.call_history, (field, args)]
            )

            if field.list:
                data = await obj._fetch_async()
                fields = obj._fields()
                remote_objects = []

                for remote_object_data in data:
                    remote_object = GraphQLRemoteObject(
                        executor=self.executor,
                        api=self.api,
                        python_type=python_type,
                        mappers=self.mappers,
                        call_history=[*self.call_history, (field, args)]
                    )

                    for field, args in fields:
                        field_value = remote_object_data.get(
                            to_camel_case(field.name)
                        )
                        arg_hash = self.hash(args)
                        field_key = (field, arg_hash)
                        remote_object.values[field_key] = field_value

                    remote_objects.append(remote_object)
                return remote_objects

            else:
                if field.mutable or field.nullable:
                    try:
                        await obj.fetch_async()
                    except NullResponse:
                        return None

                if field.mutable:
                    meta = self.mappers.mutable_mapper.meta.get(
                        (self.graphql_mutable_type.name, field.name)
                    )

                    if field.recursive and \
                            meta and \
                            meta.get(GraphQLMetaKey.resolve_to_self, True):
                        self.values.update(obj.values)
                        return self

                return obj

    def get_value(self, field: 'GraphQLRemoteField', args: Dict):
        self._map()
        value, result, arg_hash = self._get_value_cached(field, args)

        if result:
            return value

        if (field, arg_hash) in self.values.keys():
            return self.values.get((field, arg_hash), None)

        self._get_value_check_mutated(field)

        if field.scalar:
            self.fetch(fields=[(field, args)])
            return self.values.get((field, arg_hash), None)

        else:
            python_type = self.mappers.map(
                field.graphql_field.type,
                reverse=True
            )

            obj = GraphQLRemoteObject(
                executor=self.executor,
                api=self.api,
                python_type=python_type,
                mappers=self.mappers,
                call_history=[*self.call_history, (field, args)]
            )

            if field.list:
                data = obj._fetch()
                fields = obj._fields()
                remote_objects = []

                for remote_object_data in data:
                    remote_object = GraphQLRemoteObject(
                        executor=self.executor,
                        api=self.api,
                        python_type=python_type,
                        mappers=self.mappers,
                        call_history=[*self.call_history, (field, args)]
                    )

                    for field, args in fields:
                        field_value = remote_object_data.get(
                            to_camel_case(field.name)
                        )
                        arg_hash = self.hash(args)
                        field_key = (field, arg_hash)
                        remote_object.values[field_key] = field_value

                    remote_objects.append(remote_object)
                return remote_objects

            else:
                if field.mutable or field.nullable:
                    try:
                        obj.fetch()
                    except NullResponse:
                        return None

                if field.mutable:
                    meta = self.mappers.mutable_mapper.meta.get(
                        (self.graphql_mutable_type.name, field.name)
                    )

                    if field.recursive and \
                            meta and \
                            meta.get(GraphQLMetaKey.resolve_to_self, True):
                        self.values.update(obj.values)
                        return self

                return obj

    def get_field(self, name):
        self._map()

        camel_name = to_camel_case(name)
        field = None
        mutable = False

        try:
            field = self.graphql_query_type.fields.get(camel_name)
        except AssertionError:
            pass

        if field is None:
            try:
                field = self.graphql_mutable_type.fields.get(camel_name)
                mutable = True
            except AssertionError:
                pass

        if not field:
            raise GraphQLError(f"Field {name} on {self} does not exist")

        return GraphQLRemoteField(
            name=camel_name,
            mutable=mutable,
            graphql_field=field,
            parent=self
        )

    def __getattr__(self, name):
        if name == "__await__":
            raise AttributeError('Not Awaitable')

        field, auto_call = self.getattr(name)

        if auto_call:
            return field()

        return field

    async def call_async(self, name, *args, **kwargs):
        field, auto_call = self.getattr(name, pass_through=False)
        return await field.call_async(*args, **kwargs)

    def getattr(self, name, pass_through=True):
        self._map()

        attribute_type = getattr(self.python_type, name, None)

        is_dataclass_field = False

        try:
            from dataclasses import fields, is_dataclass

            if is_dataclass(self.python_type):
                # noinspection PyDataclass
                field_names = [
                    field.name
                    for field in fields(self.python_type)
                ]

                is_dataclass_field = name in field_names

        except ImportError:
            pass

        is_property = isinstance(attribute_type, property)
        is_callable = callable(attribute_type)

        auto_call = is_dataclass_field or is_property

        if not auto_call:
            try:
                from sqlalchemy.orm.attributes import InstrumentedAttribute
                auto_call = isinstance(attribute_type, InstrumentedAttribute)
            except ImportError:
                pass

        try:
            field = self.get_field(name)

        except GraphQLError as err:
            if not pass_through:
                raise err

            if "does not exist" in err.message:
                if is_callable:
                    func = getattr(self.python_type, name)
                    _is_method = inspect.ismethod(func)
                    _is_static_method = is_static_method(
                        self.python_type,
                        name
                    )

                    if _is_method or _is_static_method:
                        return func, False
                    else:
                        return (lambda *args, **kwargs: func(
                            self,
                            *args, **kwargs
                        )), False

                if is_property:
                    prop = getattr(self.python_type, name)
                    return prop.fget(self), False
            raise

        return field, auto_call

    def __str__(self):
        self._map()

        return f"<RemoteObject({self.graphql_query_type.name}) " \
               f"at {hex(id(self))}>"


class GraphQLRemoteField:

    # noinspection PyProtectedMember
    def __init__(
        self,
        name: str,
        mutable: bool,
        graphql_field: GraphQLField,
        parent: GraphQLRemoteObject
    ):
        self.name = name
        self.mutable = mutable
        self.graphql_field = graphql_field
        self.parent = parent
        self.nullable = is_nullable(self.graphql_field.type)
        self.scalar = is_scalar(self.graphql_field.type)
        self.list = is_list(self.graphql_field.type)

        self.recursive = self.parent.python_type == self.parent.mappers.map(
            self.graphql_field.type,
            reverse=True
        )

    def graphql_type(self) -> GraphQLType:
        graphql_type = self.graphql_field.type
        while hasattr(graphql_type, 'of_type'):
            graphql_type = graphql_type.of_type

        return graphql_type

    def remap_args_to_kwargs(self, args, kwargs):
        arg_names = list(self.graphql_field.args.keys())
        arg_names_count = len(arg_names)
        arg_count = len(args)

        if arg_count > arg_names_count:
            raise TypeError(
                f"{self.name} takes {arg_names_count} "
                f"argument{'s' if arg_names_count > 1 else ''} "
                f"({arg_count} given)"
            )

        for arg_index in range(0, arg_count):
            arg_name = arg_names[arg_index]
            kwargs[arg_name] = args[arg_index]

    def __call__(self, *args, **kwargs):
        if args:
            self.remap_args_to_kwargs(args=args, kwargs=kwargs)
        return self.parent.get_value(self, kwargs)

    async def call_async(self, *args, **kwargs):
        if args:
            self.remap_args_to_kwargs(args=args, kwargs=kwargs)
        return await self.parent.get_value_async(self, kwargs)

    def __hash__(self):
        return hash(hash(self.parent.python_type.__name__) + hash(self.name))

    def __eq__(self, other):
        if isinstance(other, GraphQLRemoteField):
            if other.parent == self.parent and other.name == self.name:
                return True


class GraphQLRemoteQueryBuilder:

    def __init__(
        self,
        call_stack: List[Tuple['GraphQLRemoteField', Dict]],
        fields: List[Tuple['GraphQLRemoteField', Dict]],
        mappers: GraphQLMappers,
        mutable=False
    ):
        self.call_stack = call_stack
        self.fields = fields
        self.mappers = mappers
        self.mutable = mutable

    def build(self):
        if self.mutable:
            query = 'mutation'
        else:
            query = 'query'

        def to_field_call(field, args=None):
            name = field.name
            field_call = to_camel_case(name)
            if args:
                values = []
                for key, value in args.items():
                    camel_key = to_camel_case(key)
                    graphql_arg = field.graphql_field.args[camel_key]
                    graphql_type = graphql_arg.type

                    str_value = self.map_to_input_value(
                        value=value,
                        expected_graphql_type=graphql_type,
                        mappers=self.mappers
                    )

                    if str_value is not None:
                        values.append(f"{camel_key}:{str_value}")

                field_call += f"({','.join(values)})"
            return field_call

        for field, args in self.call_stack:
            query += "{" + to_field_call(field, args=args)

        field_calls = [
            to_field_call(field, args=args)
            for field, args in self.fields
        ]

        query += "{" + ",".join(field_calls) + "}"
        query += "}" * len(self.call_stack)

        return query

    # noinspection PyMethodMayBeStatic
    def map_to_input_value(
        self,
        value,
        mappers: GraphQLMappers,
        expected_graphql_type=None
    ):
        from graphql_api.mapper import is_scalar

        if value is None:
            return None

        python_type = type(value)

        if is_scalar(python_type):
            if isinstance(value, (list, set)):
                values = [
                    self.map_to_input_value(
                        item,
                        mappers=mappers,
                        expected_graphql_type=expected_graphql_type
                    )
                    for item in value
                ]
                return '[' + ','.join(values) + ']'

            if isinstance(value, str):
                return json.dumps(value)
            if isinstance(value, bool):
                return 'true' if value else 'false'
            if isinstance(value, (float, int)):
                return str(value)
            else:
                return '"' + str(value) + '"'

        if isinstance(value, enum.Enum):
            return str(value.value)

        if isinstance(value, object):

            if expected_graphql_type is not None:

                while hasattr(expected_graphql_type, 'of_type'):
                    expected_graphql_type = expected_graphql_type.of_type

                graphql_type = expected_graphql_type
            else:
                graphql_type = mappers.query_mapper.input_type_mapper.map(
                    type(value)
                )

            input_dict = {}

            for key, field in graphql_type.fields.items():
                try:
                    raw_input_value = getattr(value, to_snake_case(key))

                    if inspect.ismethod(raw_input_value):
                        raw_input_value = raw_input_value()

                except AttributeError:
                    if not is_nullable(field.type):
                        raise GraphQLError(
                            f"InputObject error, '{type(value)}' object has"
                            f" no attribute {to_snake_case(key)}, nested"
                            f" inputs must have matching attribute "
                            f"to field names"
                        )
                else:
                    _value = self.map_to_input_value(
                        raw_input_value,
                        mappers=mappers
                    )

                    if _value is not None:
                        input_dict[key] = _value

            input_values = [
                f"{key}:{value}"
                for key, value in input_dict.items()
            ]

            input_value = "{" + ",".join(input_values) + "}"

            return input_value


def remote_execute(executor: GraphQLBaseExecutor, context):
    operation = context.request.info.operation.operation
    query = context.field.query
    redirected_query = operation.value + " " + query

    result = executor.execute(query=redirected_query)

    if result.errors:
        raise GraphQLError(str(result.errors))

    return result.data


def is_list(graphql_type):
    while hasattr(graphql_type, 'of_type'):
        if isinstance(graphql_type, GraphQLList):
            return True
        graphql_type = graphql_type.of_type

    return False


def is_scalar(graphql_type):
    while hasattr(graphql_type, 'of_type'):
        graphql_type = graphql_type.of_type

    if isinstance(graphql_type, GraphQLScalarType):
        return True

    if isinstance(graphql_type, GraphQLEnumType):
        return True

    return False


def is_nullable(graphql_type):
    while hasattr(graphql_type, 'of_type'):
        if isinstance(graphql_type, GraphQLNonNull):
            return False
        graphql_type = graphql_type.of_type

    return True


def is_static_method(klass, attr, value=None):
    if value is None:
        value = getattr(klass, attr)
    assert getattr(klass, attr) == value

    for cls in inspect.getmro(klass):
        if inspect.isroutine(value):
            if attr in cls.__dict__:
                bound_value = cls.__dict__[attr]
                if isinstance(bound_value, staticmethod):
                    return True

    return False


def to_ast_value(value, graphql_type):
    if value is None:
        return None

    type_map = {
        (bool,): ast.BooleanValueNode,
        (str,): ast.StringValueNode,
        (float,): ast.FloatValueNode,
        (int,): ast.IntValueNode
    }
    ast_type = None
    ast_value = None

    for types, ast_type in type_map.items():
        if isinstance(value, types):
            ast_value = ast_type(value=value)
            break

    if isinstance(graphql_type, GraphQLEnumType):
        if ast_type == ast.StringValueNode:
            ast_value = ast.EnumValueNode()
            ast_value.value = value

    if not ast_value:
        raise TypeError(
            f"Unable to map Python scalar type {type(value)} "
            f"to a valid GraphQL ast type"
        )
    else:
        return ast_value



from typing import List, Callable, Any, Type, Dict, Tuple

# noinspection PyPackageRequirements
from graphql import (
    GraphQLSchema,
    GraphQLObjectType,
    GraphQLField,
    GraphQLString,
    is_named_type,
    ExecutionResult,
    GraphQLType
)

from graphql_api import GraphQLError

from graphql_api.executor import GraphQLExecutor, GraphQLBaseExecutor
from graphql_api.context import GraphQLContext
from graphql_api.reduce import GraphQLSchemaReducer, GraphQLFilter
from graphql_api.mapper import GraphQLTypeMapper


class GraphQLFieldContext:

    def __init__(self, meta, query=None):
        self.meta = meta
        self.query = query

    def __str__(self):
        query_str = ""
        if self.query:
            query_str = f', query: {query_str}' if self.query else ''
        return f"<Node meta: {self.meta}{query_str}>"


class GraphQLRequestContext:

    def __init__(self, args, info):
        self.args = args
        self.info = info


def decorate(
    func: Callable,
    _type: str,
    schema: "GraphQLAPI" = None,
    meta: Dict = None
):
    func.graphql = True
    func.defined_on = func

    if not meta:
        meta = {}

    api = {
        "defined_on": func,
        "meta": meta,
        "type": _type,
        "schema": schema
    }

    if not hasattr(func, "schemas"):
        func.schemas = {}

    if hasattr(func, "schemas"):
        func.schemas[schema] = api

    return func


def decorator(a, b, _type):
    func = a if callable(a) else b if callable(b) else None
    meta = a if isinstance(a, dict) else b if isinstance(b, dict) else None
    schema = a if isinstance(a, GraphQLAPI) else \
        b if isinstance(b, GraphQLAPI) else None

    if func:
        return decorate(
            func=func,
            _type=_type,
            schema=schema,
            meta=meta
        )

    return lambda _func: decorate(
        func=_func,
        _type=_type,
        schema=schema,
        meta=meta
    )


class GraphQLAPI(GraphQLBaseExecutor):

    def field(self=None, meta=None, mutable=False):
        _type = "query"
        if mutable:
            _type = "mutation"

        return decorator(self, meta, _type=_type)

    def type(
        self=None,
        meta=None,
        abstract=False,
        interface=False,
        root=False
    ):
        _type = "object"
        if interface:
            _type = "interface"
        elif abstract:
            _type = "abstract"
        elif root:
            return self.set_root

        return decorator(self, meta, _type=_type)

    def set_root(self, root_type):
        self.root_type = root_type
        return root_type

    def __init__(
        self,
        root: Type = None,
        middleware: List[Callable[[Callable, GraphQLContext], Any]] = None,
        filters: List[GraphQLFilter] = None
    ):
        super().__init__()
        if middleware is None:
            middleware = []

        self.root_type = root
        self.middleware = middleware
        self.filters = filters
        self.query_mapper = None
        self.mutation_mapper = None

    def graphql_schema(self) -> Tuple[GraphQLSchema, Dict]:
        schema_args = {}
        meta = {}

        if self.root_type:
            # Create the root query
            query_mapper = GraphQLTypeMapper(schema=self)
            query: GraphQLType = query_mapper.map(self.root_type)

            if not isinstance(query, GraphQLObjectType):
                raise GraphQLError(
                    f"Query {query} was not a valid ObjectType."
                )

            # Filter the root query
            filtered_query = GraphQLSchemaReducer.reduce_query(
                query_mapper,
                query,
                filters=self.filters
            )

            if query_mapper.validate(filtered_query, evaluate=True):
                schema_args['query'] = filtered_query
                query_types = query_mapper.types()
                registry = query_mapper.registry

            else:
                query_types = set()
                registry = None

            # Create the root mutation
            mutation_mapper = GraphQLTypeMapper(
                as_mutable=True,
                suffix="Mutable",
                registry=registry,
                schema=self
            )
            mutation: GraphQLType = mutation_mapper.map(self.root_type)

            if not isinstance(mutation, GraphQLObjectType):
                raise GraphQLError(
                    f"Mutation {mutation} was not a valid ObjectType."
                )

            # Filter the root mutation
            filtered_mutation = GraphQLSchemaReducer.reduce_mutation(
                mutation_mapper,
                mutation
            )

            if mutation_mapper.validate(filtered_mutation, evaluate=True):
                schema_args['mutation'] = filtered_mutation
                mutation_types = mutation_mapper.types()
            else:
                mutation_types = set()

            schema_args['types'] = list(query_types | mutation_types)
            schema_args['types'] = [
                type_
                for type_ in schema_args['types'] if is_named_type(type_)
            ]

            meta = {**query_mapper.meta, **mutation_mapper.meta}

            self.query_mapper = query_mapper
            self.mutation_mapper = mutation_mapper

        # Create a placeholder query (every GraphQL schema must have a query)
        if 'query' not in schema_args:
            placeholder = GraphQLField(
                type_=GraphQLString,
                resolve=lambda *_: ''
            )
            schema_args['query'] = GraphQLObjectType(
                name='PlaceholderQuery',
                fields={'placeholder': placeholder}
            )

        schema = GraphQLSchema(**schema_args)

        return schema, meta

    def execute(
        self,
        query,
        variables=None,
        operation_name=None
    ) -> ExecutionResult:
        return self.executor().execute(
            query=query,
            variables=variables,
            operation_name=operation_name
        )

    def executor(
        self,
        root_value: Any = None,
        middleware: List[Callable[[Callable, GraphQLContext], Any]] = None,
        middleware_on_introspection: bool = False
    ) -> GraphQLExecutor:
        schema, meta = self.graphql_schema()

        if callable(self.root_type) and root_value is None:
            root_value = self.root_type()

        return GraphQLExecutor(
            schema=schema,
            meta=meta,
            root_value=root_value,
            middleware=middleware,
            middleware_on_introspection=middleware_on_introspection
        )
